/******************************************************************************
 * (c) Copyright 2001, Roger Hardiman
 * (c) Copyright 2000, Devin Butterfield, Robert A. Buaas
 *
 *	ixj_freebsd.c
 *
 *	FreeBSD Specific Core Device Driver Functions.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 *      	This file provides the FreeBSD system specific core
 *      	device driver functions for the Quicknet Internet 
 *      	PhoneJack, Internet LineJack, and PhoneCARD telephony
 *		cards. This file was derived from the original work by
 *		Ed Okerson and Greg Herlein. --DB 09/13/2000
 *
 *    	Authors:	
 *		Devin Butterfield <dbutter@wireless.net>
 *      	Robert A. Buaas <buaas@wireless.net>
 *
 *	Notes:
 *
 * 	Fixes:
 *		10/16/2000:	Merged changes from latest Linux driver
 *				release version 0.4.1 to bring code current.
 *		10/31/2000:	Fixed bug that caused ohphone to hang when
 *				user attempted to send DTMF OOB. 
 *
 *
 * $Id: ixj_freebsd.c,v 1.11 2001/08/08 16:23:50 rogerh Exp $
 *
 *****************************************************************************/
static char ixj_freebsd_c_revision[] = "$Revision: 1.11 $";

#include <sys/types.h>			/* Defines for signed/unsigned	*/
#include <sys/param.h>                  /* priorities, min(), etc.	*/
#include <sys/conf.h>			/* cdevsw stuff			*/
#include <sys/kernel.h>                 /* provides ticks, sysinit, etc.*/
#include <sys/module.h>			/* kernel module support	*/
#include <sys/systm.h>			/* for sprintf, etc.		*/
#include <sys/malloc.h>			/* memory allocation		*/
#include <sys/poll.h>			/* poll event definitions	*/
#include <sys/select.h>			/* select routines		*/
#include <sys/uio.h>                    /* moving data from/to kernel	*/
#include <sys/bus.h>                    /* bus definitions/routines	*/
#include <sys/ioccom.h>			/* ioctl macros/definitions	*/
#include <sys/fcntl.h>			/* definitions for file modes	*/
#include <machine/clock.h>		/* For DELAY()			*/
#include <machine/bus.h>		/* for inb(), outb, etc.	*/
#include <machine/resource.h>           /* resource type definitions	*/
#include <machine/ipl.h>		/* for cpl intr mask/ipending	*/
#include <sys/signalvar.h>		/* for async io / signals	*/
#include <sys/filio.h>			/* FIOASYNC, FIONBIO, defs	*/
#include <sys/rman.h>                   /* resource manager routines	*/
#include <isa/isavar.h>			/* more ISA stuff...		*/
#include "ixj.h"
#include "ixjvar_bsd.h"			/* softc, vars, prototypes, etc.*/

#define CDEV_MAJOR 	155		/* major device number		*/
#define UNIT(dev) 	(minor(dev) & 0xf)

static int	ixj_read	__P((dev_t, struct uio *, int));
static int	ixj_write	__P((dev_t, struct uio *, int));
static int	ixj_open	__P((dev_t, int, int, struct proc *));
static int	ixj_close	__P((dev_t, int, int, struct proc *));
static int	ixj_enhanced_read	__P((dev_t, struct uio *, int));
static int	ixj_enhanced_write	__P((dev_t, struct uio *, int));
static int	ixj_ioctl	__P((dev_t, u_long, caddr_t, int, struct proc *));
static int	ixj_poll	__P((dev_t, int, struct proc *));

static struct cdevsw ixj_cdevsw = {     /* character device switch	*/
	ixj_open,			/* open				*/
	ixj_close,                     	/* close			*/
	ixj_enhanced_read,            	/* read				*/
	ixj_enhanced_write,            	/* write			*/
	ixj_ioctl,                     	/* ioctl			*/
	ixj_poll,                       /* poll				*/
	nommap,
	nostrategy,
	"ixj",                          /* name				*/
	CDEV_MAJOR,			/* major			*/
	nodump,
	nopsize,
	0,
	-1                              /* bmaj				*/
};

					/* inline functions		*/

extern __inline__ void ixj_read_HSR(int board)
{
	ixj[board].hsr.bytes.low = inb_p(ixj[board].DSPbase + 8);
	ixj[board].hsr.bytes.high = inb_p(ixj[board].DSPbase + 9);
}

extern __inline__ int IsControlReady(int board)
{
	ixj_read_HSR(board);
	return ixj[board].hsr.bits.controlrdy ? 1 : 0;
}

extern __inline__ int IsPCControlReady(int board)
{
	ixj[board].pccr1.byte = inb_p(ixj[board].XILINXbase + 3);
	return ixj[board].pccr1.bits.crr ? 1 : 0;
}

extern __inline__ int IsStatusReady(int board)
{
	ixj_read_HSR(board);
	return ixj[board].hsr.bits.statusrdy ? 1 : 0;
}

extern __inline__ void ixj_kill_fasync(int board, IXJ_SIGEVENT event)
{
	if (ixj[board].async_queue) {
		if(ixj[board].ixj_signals[event]) {
			if(ixjdebug & 0x0100)
				printk("Sending signal for event %d\n", event);
		psignal(ixj[board].async_queue, ixj[board].ixj_signals[event]/*SIGIO*/ );	/* Send apps notice of change	*/
		}
	}
}


extern __inline__ void LED_SetState(int state, int board)
{
	if (ixj[board].cardtype == QTI_LINEJACK) {
		ixj[board].pld_scrw.bits.led1 = state & 0x1 ? 1 : 0;
		ixj[board].pld_scrw.bits.led2 = state & 0x2 ? 1 : 0;
		ixj[board].pld_scrw.bits.led3 = state & 0x4 ? 1 : 0;
		ixj[board].pld_scrw.bits.led4 = state & 0x8 ? 1 : 0;

		outb_p(ixj[board].pld_scrw.byte, ixj[board].XILINXbase);
	}
}

extern __inline__ void set_play_volume(int board, int volume)
{
	if (ixjdebug & 0x0002)
		printk(KERN_INFO "IXJ: /dev/phone%d Setting Play Volume to 0x%4.4x\n", board, volume);	
	ixj_WriteDSPCommand(0xCF02, board);
	ixj_WriteDSPCommand(volume, board);
}

extern __inline__ int get_play_volume(int board)
{
	ixj_WriteDSPCommand(0xCF00, board);
	return ixj[board].ssr.high << 8 | ixj[board].ssr.low;
}

extern __inline__ void set_play_depth(int board, int depth)
{
	if (depth > 60)
		depth = 60;
	if (depth < 0)
		depth = 0;
	ixj_WriteDSPCommand(0x5280 + depth, board);
}

extern __inline__ int get_play_level(int board)
{
	int retval;

	ixj_WriteDSPCommand(0xCF8F, board); /* 8022 Reference page 9-38 */
	return ixj[board].ssr.high << 8 | ixj[board].ssr.low;
	retval = ixj[board].ssr.high << 8 | ixj[board].ssr.low;
	retval = (retval * 256) / 240;
	return retval;
}

int
ixj_attach(int board, struct ixj_softc *sc)
{
	unsigned short cmd;
	unsigned long jif;
	int cnt;
	IXJ *j = &ixj[board];
	int error;

	if (ixjdebug & 0x0002)
		printf("Write IDLE to Software Control Register\n");
                                        /* Put DSP in full power mode. 	*/
	ixj_WriteDSPCommand(0x0FE0, board);
                                        /* write IDLE to Software	*/
					/* Control Register.		*/
	if (ixj_WriteDSPCommand(0x0000, board))
		return -1;
					/* The read values of the SSR 	*/
					/* should be 0x00 for the IDLE 	*/
					/* command.			*/
	if (j->ssr.low || j->ssr.high) {
		printf("ssr.low=0x%x, ssr.high=0x%x, should be 0x0000\n", 
				j->ssr.low, j->ssr.high);
		return -1;
	}
	
	if (ixjdebug & 0x0002)
		printf("Get Device ID Code\n");
	if (ixj_WriteDSPCommand(0x3400, board)) /* Get Device ID Code 	*/
		return -1;

	j->dsp.low = j->ssr.low;
	j->dsp.high = j->ssr.high;
	
	if(ixjdebug & 0x0002)
		printf("DSP ID code: %x\n", ((j->dsp.high << 8) 
					+ j->dsp.low));

	if (ixjdebug > 0)
		printf("Get Device Version Code\n");
	if (ixj_WriteDSPCommand(0x3800, board))/* Get Dev Version Code	*/
		return -1;

	j->ver.low = j->ssr.low;
	j->ver.high = j->ssr.high;

	if(ixjdebug > 0x0002)
		printf("DSP version code: %x\n", ((j->ver.high << 8) 
					+ j->ver.low));
	
					/* NOTE:			*/
					/* Here we assume that resources*/
					/* would have already been	*/
					/* allocated by the probe/attach*/
					/* routines for any cards found.*/
	if (j->cardtype) {
		switch (j->cardtype) {
		case QTI_PHONEJACK:	/* Internet PhoneJACK		*/
			if (j->dsp.low != 0x20) {
				j->dsp.high = 0x80;
				j->dsp.low = 0x20;
				ixj_WriteDSPCommand(0x3800, board);
				j->ver.low = j->ssr.low;
				j->ver.high = j->ssr.high;
			}
			break;
		case QTI_LINEJACK:	/* Internet LineJACK		*/
					/* we already allocated 	*/
					/* resources for the XILINX	*/
					/* so we're good to go.		*/
			break;
		case QTI_PHONEJACK_LITE:/* Internet PhoneJACK Lite...	*/
		case QTI_PHONEJACK_PCI:	/* Internet PhoneJACK PCI...	*/
					/* we already allocated 	*/
					/* resources for the XILINX	*/
					/* so we're good to go.		*/
			j->pld_slicw.pcib.e1 = 1;
			outb(j->XILINXbase, j->pld_slicw.byte); 
			break;
		case QTI_PHONECARD:	/* Internet PhoneCARD		*/
			break;
		}
	}
	if (j->dsp.low == 0x20 || j->cardtype == QTI_PHONEJACK_LITE || 
			j->cardtype == QTI_PHONEJACK_PCI) {
		if (ixjdebug & 0x0002)
			printf("Write CODEC config to Software Control Register\n");
                                        /* Write CODEC config to	*/
					/* Software Control Register. 	*/
		if (ixj_WriteDSPCommand(0xC462, board))
			return -1;
		if (ixjdebug & 0x0002)
			printf("Write CODEC timing to Software Control Register\n");
		if (j->cardtype == QTI_PHONEJACK) {
			cmd = 0x9FF2;
		} else {
			cmd = 0x9FF5;
		}                       /* Write CODEC timing to 	*/
					/* Software Control Register. 	*/
		if (ixj_WriteDSPCommand(cmd, board))
			return -1;
	} else {
		if (set_base_frame(board, 30) != 30)
			return -1;
		if (ixjdebug & 0x0002)
			printf("Write CODEC config to Software Control Register\n");
		if (j->cardtype == QTI_PHONECARD) {
                                        /* Write CODEC config to	*/
					/* Software Control Register. 	*/
			if (ixj_WriteDSPCommand(0xC528, board))
				return -1;
		}
		if (j->cardtype == QTI_LINEJACK) {
                                        /* Write CODEC config to	*/
					/* Software Control Register. 	*/
			if (ixj_WriteDSPCommand(0xC528, board))
				return -1;
			if (ixjdebug & 0x0002)
				printf("Turn on the PLD Clock at 8Khz\n");
			j->pld_clock.byte = 0;
			outb(j->XILINXbase + 0x04, j->pld_clock.byte); 	
		}
	}

	if (j->dsp.low == 0x20) {
		if (ixjdebug > 0)
			printf("Configure GPIO pins\n");
		j->gpio.bytes.high = 0x09;
/*  bytes.low = 0xEF;  0xF7 	*/
		j->gpio.bits.gpio1 = 1;
		j->gpio.bits.gpio2 = 1;
		j->gpio.bits.gpio3 = 0;
		j->gpio.bits.gpio4 = 1;
		j->gpio.bits.gpio5 = 1;
		j->gpio.bits.gpio6 = 1;
		j->gpio.bits.gpio7 = 1; /* Set GPIO pin directions 	*/
		ixj_WriteDSPCommand(j->gpio.word, board);
		if (ixjdebug > 0)
			printf("Enable SLIC\n");
		j->gpio.bytes.high = 0x0B;
		j->gpio.bytes.low = 0x00;
		j->gpio.bits.gpio1 = 0;
		j->gpio.bits.gpio2 = 1;
		j->gpio.bits.gpio5 = 0; /* send the ring stop signal 	*/
		ixj_WriteDSPCommand(j->gpio.word, board);
		j->port = PORT_POTS;
	} else {
		if (j->cardtype == QTI_LINEJACK) {
			if (ixjdebug > 0)
				printf("initializing LineJack...\n");
			LED_SetState(0x1, board);
			jif = jiffies + (hertz / 10);
			while (time_before(jiffies, jif)) {
				error = tsleep((caddr_t)&j, PZERO | PCATCH, "ixjslp", 1);
				if(error == EINTR)
					return (error);
			}
			LED_SetState(0x2, board);
			jif = jiffies + (hertz / 10);
			while (time_before(jiffies, jif)) {
				error = tsleep((caddr_t)&j, PZERO | PCATCH, "ixjslp", 1);
				if(error == EINTR)
					return (error);
			}
			LED_SetState(0x4, board);
			jif = jiffies + (hertz / 10);
			while (time_before(jiffies, jif)) {
				error = tsleep((caddr_t)&j, PZERO | PCATCH, "ixjslp", 1);
				if(error == EINTR)
					return (error);
			}
			LED_SetState(0x8, board);
			jif = jiffies + (hertz / 10);
			while (time_before(jiffies, jif)) {
				error = tsleep((caddr_t)&j, PZERO | PCATCH, "ixjslp", 1);
				if(error == EINTR)
					return (error);
			}
			LED_SetState(0x0, board);
			daa_get_version(board);
			if (ixjdebug & 0x0002)
				printf("Loading DAA Coefficients\n");
			DAA_Coeff_US(board);
			if (!ixj_daa_write(board))
				printf("DAA write failed on board %d\n", board);
			ixj_daa_cid_reset(board);
			j->flags.pots_correct = 0;
			j->flags.pstn_present = 0;
			ixj_linetest(board);
			if (j->flags.pots_correct) {
					/* Turn off DAA Frame Sync	*/
				j->pld_scrw.bits.daafsyncen = 0;
				outb(j->XILINXbase, j->pld_scrw.byte);
				j->pld_slicw.bits.rly1 = 1;
				j->pld_slicw.bits.spken = 1;
				outb(j->XILINXbase + 0x01, j->pld_slicw.byte);
				SLIC_SetState(PLD_SLIC_STATE_STANDBY, board);
				j->port = PORT_POTS;
			}
			ixj_set_port(board, PORT_PSTN);
			ixj_set_pots(board, 1);
			if (ixjdebug & 0x0002)
				printf("Enable Mixer\n");
			ixj_mixer(0x0000, board);	/*Master Volume Left unmute 0db */
			ixj_mixer(0x0100, board);	/*Master Volume Right unmute 0db */

			ixj_mixer(0x0203, board);	/*Voice Left Volume unmute 6db */
			ixj_mixer(0x0303, board);	/*Voice Right Volume unmute 6db */

			ixj_mixer(0x0480, board);	/*FM Left mute */
			ixj_mixer(0x0580, board);	/*FM Right mute */

			ixj_mixer(0x0680, board);	/*CD Left mute */
			ixj_mixer(0x0780, board);	/*CD Right mute */

			ixj_mixer(0x0880, board);	/*Line Left mute */
			ixj_mixer(0x0980, board);	/*Line Right mute */

			ixj_mixer(0x0A80, board);	/*Aux left mute  */
			ixj_mixer(0x0B80, board);	/*Aux right mute */

			ixj_mixer(0x0C00, board);	/*Mono1 unmute 12db */
			ixj_mixer(0x0D80, board);	/*Mono2 mute */

			ixj_mixer(0x0E80, board);	/*Mic mute */

			ixj_mixer(0x0F00, board);	/*Mono Out Volume unmute 0db */

			ixj_mixer(0x1000, board);	/*Voice Left and Right out only */
			ixj_mixer(0x110C, board);


			ixj_mixer(0x1200, board);	/*Mono1 switch on mixer left */
			ixj_mixer(0x1401, board);

			ixj_mixer(0x1300, board);       /*Mono1 switch on mixer right */
			ixj_mixer(0x1501, board);

			ixj_mixer(0x1700, board);	/*Clock select */

			ixj_mixer(0x1800, board);	/*ADC input from mixer */

			ixj_mixer(0x1900, board);	/*Mic gain 0db */
			if (ixjdebug & 0x0002)
				printf("ixj%d: Setting Default US Ring Cadende Detection\n", board);
			j->cadence_f[4].state = 0;
			j->cadence_f[4].on1 = 0;	/* Cadence Filter 4 is used for PSTN ring cadence */
			j->cadence_f[4].off1 = 0;
			j->cadence_f[4].on2 = 0;
			j->cadence_f[4].off2 = 0;
			j->cadence_f[4].on3 = 0;
			j->cadence_f[4].off3 = 0;	/* These should represent standard US ring pulse. */
			j->pstn_last_rmr = jiffies;

		} else {
			if (j->cardtype == QTI_PHONECARD) {
				ixj_WriteDSPCommand(0xCF07, board);
				ixj_WriteDSPCommand(0x00B0, board);
				ixj_set_port(board, PORT_SPEAKER);
			} else {
				ixj_set_port(board, PORT_POTS);
				SLIC_SetState(PLD_SLIC_STATE_STANDBY, board);
			}
		}
	}

	j->intercom = -1;
	j->framesread = j->frameswritten = 0;
	j->read_wait = j->write_wait = 0;
	j->rxreadycheck = j->txreadycheck = 0;
	set_play_volume(board, 0x100);
	set_rec_volume(board, 0x100);
					/* Write IDLE to Software 	*/
					/* Control Register. 		*/
	if (ixj_WriteDSPCommand(0x0000, board))
		return -1;
					/* The read values of the SSR	*/
					/* should be 0x00 for the IDLE 	*/
					/* command.			*/
	if (j->ssr.low || j->ssr.high)
		return -1;
	if (ixjdebug & 0x0002)
		printf("Enable Line Monitor\n");
	if (ixjdebug & 0x0002)
		printf("Set Line Monitor to Asyncronous Mode\n");
	if (ixj_WriteDSPCommand(0x7E01, board))	/* Async Line Monitor	*/
		return -1;

	if (ixjdebug & 0x0002)
		printf("Enable DTMF Detectors\n");

	if (ixj_WriteDSPCommand(0x5151, board))	/* Enable DTMF detection*/
		return -1;

	if (ixj_WriteDSPCommand(0x6E01, board))	/* Set Async Tone Gener.*/
		return -1;

	set_rec_depth(board, 2);	/* Set Rec Chan Lim to 2 frames	*/

	set_play_depth(board, 2);	/* Set Plbk Chan Lim to 2 frames*/

	j->ex.bits.dtmf_ready = 0;
	j->dtmf_state = 0;
	j->dtmf_wp = j->dtmf_rp = 0;
	j->rec_mode = j->play_mode = -1;
	j->flags.ringing = 0;
	j->maxrings = MAXRINGS;
	j->ring_cadence = USA_RING_CADENCE;
	j->drybuffer = 0;
	j->winktime = 320;
	j->flags.dtmf_oob = 0;
	for (cnt = 0; cnt < 4; cnt++)
		j->cadence_f[cnt].enable = 0;
					/* must be a device on the 	*/
					/* specified address. 		*/
	j->board = board;
	ixj_WriteDSPCommand(0x0FE3, board);/* Put DSP in 1/5 pwr mode.	*/

	/* Set up the default signals for events */
	for (cnt = 0; cnt < 35; cnt++)
		j->ixj_signals[cnt] = SIGIO;

	/* Set the excetion signal enable flags */
	j->ex_sig.bits.dtmf_ready = j->ex_sig.bits.hookstate = j->ex_sig.bits.flash = j->ex_sig.bits.pstn_ring = 
	j->ex_sig.bits.caller_id = j->ex_sig.bits.pstn_wink = j->ex_sig.bits.f0 = j->ex_sig.bits.f1 = j->ex_sig.bits.f2 = 
	j->ex_sig.bits.f3 = j->ex_sig.bits.fc0 = j->ex_sig.bits.fc1 = j->ex_sig.bits.fc2 = j->ex_sig.bits.fc3 = 1;
	j->fskdata = NULL;
	j->cidcw_wait = 0;
					/* make the device entry	*/
     	sc->dev = make_dev(&ixj_cdevsw, board, UID_ROOT,
					GID_WHEEL, 0666, "phone%d", board);
	add_caps(board);
	return (0);
}

static  int
ixj_open(dev_t dev, int flags, int mode, struct proc *p)
{
	int unit = UNIT(dev);
	int s;
        IXJ *j = &ixj[unit];

	if (!j->DSPbase) {		/* make sure device is config'd	*/
		printf("***Cannot open phone device, no i/o base address!\n");
		return ENODEV;
	}

	if (ixjdebug > 0)
		printf("Opening board %d\n", unit);

	switch (flags & (FREAD | FWRITE)) {
	case FREAD:                     /* opened for read?		*/
		if(!j->readers) {	/* if not already readers...	*/
			j->readers++;   /* add reader.			*/
		} else {
			return EBUSY;   /* if already readers, ret busy.*/
		}
		break;
	case FWRITE:			/* opened for write?		*/
		if(!j->writers) {       /* if not already writers...	*/
			j->writers++;   /* add writer.			*/
		} else {
			return EBUSY;   /* if already writers, ret busy.*/
		}	
		break;
	case (FREAD | FWRITE):		/* opened for read-write?	*/
		if(!j->readers && !j->writers) {/* if not already RW's..*/
			j->readers++;   /* add reader...		*/
			j->writers++;   /* add writer...		*/
		} else {
			return EBUSY;   /* if already RW's, ret busy.	*/
		}
		break;
	default:
		break;
	}

	s = splhigh();			/* we don't want to be		*/
					/* interrupted here.		*/
	if (j->cardtype == QTI_PHONECARD) {
		j->pslic.bits.powerdown = 0;
		j->psccr.bits.dev = 3;
		j->psccr.bits.rw = 0;
		outw(j->XILINXbase + 0x00, j->psccr.byte << 8 | j->pslic.byte); 
		ixj_PCcontrol_wait(unit);
	}
	splx(s);			/* return norm interrupt level	*/

	j->flags.cidplay = 0;
	j->flags.cidcw_ack = 0;
	j->device_state |= IXJ_STATE_OPEN; /* set dev state open flag	*/
	j->framesread = j->frameswritten = 0; /* init RDWR frames fields*/
	j->async_queue = 0;		/* initialize async queue	*/
	
	return (0);
}

static  int
ixj_close(dev_t dev, int flags, int mode, struct proc *p)
{
        IXJ_TONE ti;
	int cnt, error=0;
	int board = UNIT(dev);
        IXJ *j = &ixj[board];

	if (ixjdebug & 0x0002)
		printf("Closing board %d\n", board);
	
	/*
	 *    Set up locks to ensure that only one process is talking to the DSP at a time.
	 *    This is necessary to keep the DSP from locking up.
	 */
	while(test_and_set_bit(board, (void *)&j->busyflags) != 0) {
		error = tsleep((caddr_t)&j, PZERO | PCATCH, "ixjslp", 1);
		if(error == EINTR)
			return (error);	
	}
	if (j->cardtype == QTI_PHONECARD)
		ixj_set_port(board, PORT_SPEAKER);
	else
		ixj_set_port(board, PORT_POTS);
	aec_stop(board);
	ixj_play_stop(board);
	ixj_record_stop(board);
	set_play_volume(board, 0x100);
	set_rec_volume(board, 0x100);
	ixj_ring_off(board);
					/* Restore the tone table to 	*/
					/* default settings. 		*/
	ti.tone_index = 10;
	ti.gain0 = 1;
	ti.freq0 = hz941;
	ti.gain1 = 0;
	ti.freq1 = hz1209;
	ti.tone_index = 11;
	ti.gain0 = 1;
	ti.freq0 = hz941;
	ti.gain1 = 0;
	ti.freq1 = hz1336;
	ti.tone_index = 12;
	ti.gain0 = 1;
	ti.freq0 = hz941;
	ti.gain1 = 0;
	ti.freq1 = hz1477;
	ti.tone_index = 13;
	ti.gain0 = 1;
	ti.freq0 = hz800;
	ti.gain1 = 0;
	ti.freq1 = 0;
	ixj_init_tone(board, &ti);
	ti.tone_index = 14;
	ti.gain0 = 1;
	ti.freq0 = hz1000;
	ti.gain1 = 0;
	ti.freq1 = 0;
	ixj_init_tone(board, &ti);
	ti.tone_index = 15;
	ti.gain0 = 1;
	ti.freq0 = hz1250;
	ti.gain1 = 0;
	ti.freq1 = 0;
	ixj_init_tone(board, &ti);
	ti.tone_index = 16;
	ti.gain0 = 1;
	ti.freq0 = hz950;
	ti.gain1 = 0;
	ti.freq1 = 0;
	ixj_init_tone(board, &ti);
	ti.tone_index = 17;
	ti.gain0 = 1;
	ti.freq0 = hz1100;
	ti.gain1 = 0;
	ti.freq1 = 0;
	ixj_init_tone(board, &ti);
	ti.tone_index = 18;
	ti.gain0 = 1;
	ti.freq0 = hz1400;
	ti.gain1 = 0;
	ti.freq1 = 0;
	ixj_init_tone(board, &ti);
	ti.tone_index = 19;
	ti.gain0 = 1;
	ti.freq0 = hz1500;
	ti.gain1 = 0;
	ti.freq1 = 0;
	ixj_init_tone(board, &ti);
	ti.tone_index = 20;
	ti.gain0 = 1;
	ti.freq0 = hz1600;
	ti.gain1 = 0;
	ti.freq1 = 0;
	ixj_init_tone(board, &ti);
	ti.tone_index = 21;
	ti.gain0 = 1;
	ti.freq0 = hz1800;
	ti.gain1 = 0;
	ti.freq1 = 0;
	ixj_init_tone(board, &ti);
	ti.tone_index = 22;
	ti.gain0 = 1;
	ti.freq0 = hz2100;
	ti.gain1 = 0;
	ti.freq1 = 0;
	ixj_init_tone(board, &ti);
	ti.tone_index = 23;
	ti.gain0 = 1;
	ti.freq0 = hz1300;
	ti.gain1 = 0;
	ti.freq1 = 0;
	ixj_init_tone(board, &ti);
	ti.tone_index = 24;
	ti.gain0 = 1;
	ti.freq0 = hz2450;
	ti.gain1 = 0;
	ti.freq1 = 0;
	ixj_init_tone(board, &ti);
	ti.tone_index = 25;
	ti.gain0 = 1;
	ti.freq0 = hz350;
	ti.gain1 = 0;
	ti.freq1 = hz440;
	ixj_init_tone(board, &ti);
	ti.tone_index = 26;
	ti.gain0 = 1;
	ti.freq0 = hz440;
	ti.gain1 = 0;
	ti.freq1 = hz480;
	ixj_init_tone(board, &ti);
	ti.tone_index = 27;
	ti.gain0 = 1;
	ti.freq0 = hz480;
	ti.gain1 = 0;
	ti.freq1 = hz620;
	ixj_init_tone(board, &ti);

	
	set_rec_depth(board, 2);	/* Set Record Channel Limit to 2 frames */

	set_play_depth(board, 2);	/* Set Playback Channel Limit to 2 frames */

	j->ex.bits.dtmf_ready = 0;
	j->dtmf_state = 0;
	j->dtmf_wp = j->dtmf_rp = 0;
	j->rec_mode = j->play_mode = -1;
	j->flags.ringing = 0;
	j->maxrings = MAXRINGS;
	j->ring_cadence = USA_RING_CADENCE;
	if(j->cadence_f[5].enable) {
		j->cadence_f[5].enable = j->cadence_f[5].state = 0;
	}	
	j->drybuffer = 0;
	j->winktime = 320;
	j->flags.dtmf_oob = 0;
	for (cnt = 0; cnt < 4; cnt++)
		j->cadence_f[cnt].enable = 0;
	idle(board);
	if(j->cardtype == QTI_PHONECARD) {
		SLIC_SetState(PLD_SLIC_STATE_OC, board);
	}	

     	switch (flags & (FREAD | FWRITE)) {
	case FREAD:                     /* was opened for read?		*/
		j->readers--;   	/* remove reader.		*/
		break;
	case FWRITE:			/* was opened for write?	*/
		j->writers--;   	/* remove writer.		*/
		break;
	case (FREAD | FWRITE):		/* was opened for read-write?	*/
		j->readers--;   	/* remove reader...		*/
		j->writers--;   	/* remove writer...		*/
		break;
	default:
		break;
	}

	if (j->read_buffer && !j->readers) {
		free(j->read_buffer, M_DEVBUF);
		j->read_buffer = NULL;
		j->read_buffer_size = 0;
	}
	if (j->write_buffer && !j->writers) {/* XXX should set write_buffer_wp=write_buffer_rp=NULL !!*/
		free(j->write_buffer, M_DEVBUF);
		j->write_buffer = NULL;
		j->write_buffer_size = 0;
	}
	j->rec_codec = j->play_codec = 0;
	j->rec_frame_size = j->play_frame_size = 0;
	j->flags.cidsent = j->flags.cidring = 0;

	if(j->cardtype == QTI_LINEJACK && !j->readers && !j->writers) {
		ixj_set_port(board, PORT_PSTN);
		daa_set_mode(board, SOP_PU_SLEEP);
		ixj_set_pots(board, 1);
	}
	ixj_WriteDSPCommand(0x0FE3, board);/* Put DSP in 1/5 power mode.*/

	j->device_state = 0x0000;	/* unset open flag		*/
	j->async_queue = 0;		/* remove procs from asyncio	*/
	funsetown(j->ixj_sigio);	/* remove F_SETOWN		*/

	/* Set up the default signals for events */
	for (cnt = 0; cnt < 35; cnt++)
		j->ixj_signals[cnt] = SIGIO;

	/* Set the excetion signal enable flags */
	j->ex_sig.bits.dtmf_ready = j->ex_sig.bits.hookstate = j->ex_sig.bits.flash = j->ex_sig.bits.pstn_ring = 
	j->ex_sig.bits.caller_id = j->ex_sig.bits.pstn_wink = j->ex_sig.bits.f0 = j->ex_sig.bits.f1 = j->ex_sig.bits.f2 = 
	j->ex_sig.bits.f3 = j->ex_sig.bits.fc0 = j->ex_sig.bits.fc1 = j->ex_sig.bits.fc2 = j->ex_sig.bits.fc3 = 1;
	clear_bit(board, &j->busyflags);
	return (0);
}

static int
ixj_enhanced_read(dev_t dev, struct uio *uio, int flags)
{
	int pre_retval;
	int retval = 0;
	int read_retval = 0;
	IXJ *j = &ixj[UNIT(dev)];

	pre_retval = ixj_PreRead(j, 0L);
	switch (pre_retval) {
	case NORMAL:
		read_retval = ixj_read(dev, uio, flags);
		ixj_PostRead(j, 0L);
		break;
	case NOPOST:
		read_retval = ixj_read(dev, uio, flags);
		break;
	case POSTONLY:
		ixj_PostRead(j, 0L);
		break;
	default:
		read_retval = pre_retval;
		break;
	}
	
	if (read_retval > 0) {
		retval = 0;
	} else 
		retval = -1;
	return retval;
}

static  int
ixj_read(dev_t dev, struct uio *uio, int flags)
{
	int unit = UNIT(dev);
	IXJ *j = &ixj[unit];
	int error = 0;
	int sz = min(uio->uio_resid, j->read_buffer_size);

	if (j->flags.inread)
		return -EALREADY;
	j->flags.inread = 1;

	while (!j->read_buffer_ready || (j->dtmf_state && j->flags.dtmf_oob)) {
		++j->read_wait;
		if (flags & O_NONBLOCK) {
			j->flags.inread = 0;
			return (-EWOULDBLOCK);
		}
		if (!ixj_hookstate(unit)) {
			j->flags.inread = 0;
			return 0;
		}
		error = tsleep(&j->read_q, PZERO | PCATCH, "ixjrd", 0);
		if (error == EINTR) {
			j->flags.inread = 0;
			return -EINTR;
		}
	}
	if(j->rec_codec == ALAW)
		ulaw2alaw(j->read_buffer, min(sz, j->read_buffer_size));	
					/* Don't ever copy more than 	*/
					/* the user asks.		*/
	error = uiomove(j->read_buffer, sz, uio);
	j->read_buffer_ready = 0;
	if (error) {
		j->flags.inread = 0;
		return -EFAULT;
	} else {
		j->flags.inread = 0;
		return sz;
	}

}

static int
ixj_enhanced_write(dev_t dev, struct uio *uio, int flags)
{
	int pre_retval;
	int retval = 0;
	ssize_t write_retval = 0;
	IXJ *j = &ixj[UNIT(dev)];

	pre_retval = ixj_PreWrite(j, 0L);
	switch (pre_retval) {
	case NORMAL:
		write_retval = ixj_write(dev, uio, flags);
		if (write_retval > 0) {
			ixj_PostWrite(j, 0L);
			j->write_buffer_wp += write_retval;
			j->write_buffers_empty--;
			retval = 0;
		}
		break;
	case NOPOST:
		write_retval = ixj_write(dev, uio, flags);
		if (write_retval > 0) {
			j->write_buffer_wp += write_retval;
			j->write_buffers_empty--;
			retval = 0;
		}
		break;
	case POSTONLY:
		ixj_PostWrite(j, 0L);
		break;
	default:
		write_retval = pre_retval;
	}
	return retval;
}

static  int
ixj_write(dev_t dev, struct uio *uio, int flags)
{
	int unit = UNIT(dev);
	IXJ *j = &ixj[unit];
        int error = 0;
	int sz = min(uio->uio_resid, j->write_buffer_size);

	if (j->flags.inwrite)
		return -EALREADY;

	j->flags.inwrite = 1;
	
	while (!j->write_buffers_empty) {
		++j->write_wait;
		if (flags & O_NONBLOCK) {
			j->flags.inwrite = 0;
			return (-EAGAIN);
		}
		if (!ixj_hookstate(unit)) {
			j->flags.inwrite = 0;
			return 0;
		}
		error = tsleep(&j->write_q, PZERO | PCATCH, "ixjwr", 0);
		if (error) {
			j->flags.inwrite = 0;
			return -EINTR;
		}
	}
	if (j->write_buffer_wp + sz >= j->write_buffer_end)
		j->write_buffer_wp = j->write_buffer;
	error = uiomove(j->write_buffer_wp, sz, uio);
	if (error) {
		j->flags.inwrite = 0;
		return -EFAULT;
	}
        if(j->play_codec == ALAW)
               alaw2ulaw(j->write_buffer_wp, min(sz, j->write_buffer_size));
	j->flags.inwrite = 0;
	return sz;
}

static int
ixj_ioctl (dev_t dev, u_long cmd, caddr_t arg, int flag, struct proc *p)
{
	IXJ_TONE ti;
	IXJ_FILTER jf;
	unsigned int mynor = minor(dev);
	unsigned int mant, raise;
	int board = UNIT(dev);
	int error=0;
	IXJ *j = &ixj[board];

	int retval = 0;

	/*
	 *    Set up locks to ensure that only one process is talking to the DSP at a time.
	 *    This is necessary to keep the DSP from locking up.
	 */
	while(test_and_set_bit(board, (void *)&j->busyflags) != 0) {
		error = tsleep((caddr_t)&j, PZERO | PCATCH, "ixjslp", 1);
		if(error == EINTR)
			return (error);	
	}
	
	if(ixjdebug > 0x0040)
		printf("phone%d ioctl, cmd: 0x%lx, arg:0x%x\n", mynor, cmd, *arg);
	
	if (mynor >= IXJMAX) {
		clear_bit(board, &j->busyflags);
		return ENODEV;
	}

	if (suser(p)) {			/* Check if non-suser tried to	*/
		switch (cmd) {          /* call ioctl that only suser	*/
		case IXJCTL_TESTRAM:    /* can call.			*/
		case IXJCTL_HZ:
			return EPERM;
		}
	}

	switch (cmd) {
	case FIOASYNC:			/* add process that wants SIGIO	*/
		if (*(int *)arg) {      /* to async queue.		*/
			if(j->async_queue != NULL) {
				return EBUSY;
			}
			j->async_queue = p;
		} else {
			j->async_queue = NULL;
		}
		break;
	case FIONBIO:			/* std ioctl called by kernel	*/
		break;
	case FIOSETOWN:                 /* F_SETOWN ioctl 		*/
					/* taken from /sys/net/if_tun.c	*/
		return (fsetown(*(int *)arg, &j->ixj_sigio));
	case FIOGETOWN:			/* F_GETOWN ioctl		*/
					/* taken from /sys/net/if_tun.c	*/
		*(int *)arg = fgetown(j->ixj_sigio);
		return (0);
	case IXJCTL_TESTRAM:		/* do card selftest...		*/
		ixj_testram(board);
		retval = (j->ssr.high << 8) + j->ssr.low;
		break;
	case IXJCTL_CARDTYPE:           /* return cardtype		*/
		retval = j->cardtype;	
		break;
	case IXJCTL_SERIAL:             /* return serial number		*/
		retval = j->serial;
		break;
	case IXJCTL_VERSION:            /* return version string	*/ 
              	memset((char *)arg, *ixj_freebsd_c_revision, sizeof(ixj_freebsd_c_revision)); /* XXX 2nd arg ! */
		break;
	case PHONE_RING_CADENCE:        /* set ring cadence		*/
		j->ring_cadence = *(short *)arg;	
		break;
	case IXJCTL_CIDCW:		/* caller id on call waiting	*/
		if(*(int *)arg) {
			j->cid_send = *(PHONE_CID *)arg;
		}
		else {
			bzero(&j->cid_send, sizeof(PHONE_CID));
		}
		ixj_write_cidcw(board);
		break;
	case PHONE_RING_START:		/* start pots port ringing	*/
		if(*(int *)arg) {       /* check if cid data is avail.	*/
			j->cid_send = *(PHONE_CID *)arg;
		} else {
			bzero(&j->cid_send, sizeof(PHONE_CID));
		}
		ixj_ring_start(board);
		break;
	case PHONE_RING_STOP:		/* stop ring cadence		*/
		j->flags.cringing = 0;
		if(j->cadence_f[5].enable) {
			j->cadence_f[5].state = 0;
		}
		ixj_ring_off(board);
		break;
	case PHONE_RING:		/* ring the pots and indicate if*/ 
					/* phone was answered or max	*/
					/* rings reached.		*/
		retval = ixj_ring(board);
		break;
	case PHONE_EXCEPTION:		/* Returns an integer with the 	*/
					/* low 14 bits holding the 	*/
					/* exception type.		*/
		retval = j->ex.bytes;
		if (j->ex.bits.flash) {
			j->flash_end = 0;
			j->ex.bits.flash = 0;
		}
		j->ex.bits.flash = 0;
		j->ex.bits.pstn_ring = 0;
		j->ex.bits.caller_id = 0;
		j->ex.bits.pstn_wink = 0;
		j->ex.bits.f0 = 0;
		j->ex.bits.f1 = 0;
		j->ex.bits.f2 = 0;
		j->ex.bits.f3 = 0;
		j->ex.bits.fc0 = 0;
		j->ex.bits.fc1 = 0;
		j->ex.bits.fc2 = 0;
		j->ex.bits.fc3 = 0;
		j->ex.bits.reserved = 0;
		break;
	case PHONE_HOOKSTATE:		/* poll pots port for hookstate	*/
					/* change.			*/
		j->ex.bits.hookstate = 0;
		retval = j->r_hook;
		break;
	case IXJCTL_SET_LED:		/* change led states on linejack*/
		LED_SetState(*(int *)arg, board);
		break;
	case PHONE_FRAME:		/* set audio frame length that	*/
					/* linejack will return.	*/
		retval = set_base_frame(board, *(int *)arg);
		break;
	case PHONE_REC_CODEC:		/* set record channel codec.	*/
					/* returns 0 if accepted.	*/
		retval = set_rec_codec(board, *(int *)arg);
		break;
	case PHONE_VAD:                	/* turn on/off voice activity	*/
					/* detection.			*/
		ixj_vad(board, *(int *)arg);
		break;
	case PHONE_REC_START:		/* activate DSP record mode.	*/
		ixj_record_start(board);
		break;
	case PHONE_REC_STOP:		/* deactivate DSP record mode.	*/
		ixj_record_stop(board);
		break;
	case PHONE_REC_DEPTH:		/* set the number of 30ms frames*/
					/* the DSP will queue. This 	*/
					/* affects audio latency. Set to*/
					/* min val of 1 or 2. to reduce	*/
					/* latency.			*/
		set_rec_depth(board, *(int *)arg);
		break;
	case PHONE_REC_VOLUME:		/* set digital scaling factor	*/
					/* for audio volume on DSP.	*/
		if(*(int *)arg == -1) { /* if -1 return current setting.*/
			retval = get_rec_volume(board);
		}
		else {                  /* otherwise, set volume.	*/
			set_rec_volume(board, *(int *)arg);
			retval = *(int *)arg;
		}
		break;
	case IXJCTL_DTMF_PRESCALE:
		if(*(int *)arg == -1) {
			retval = get_dtmf_prescale(board);
		} else {
			set_dtmf_prescale(board, *(int *)arg);
			retval = *(int *)arg;
		}
		break;		
	case PHONE_REC_LEVEL:		/* check current rec level.	*/
		retval = get_rec_level(board);
		break;
	case IXJCTL_SC_RXG:
		retval = ixj_siadc(board, *(int *)arg);
		break;
	case IXJCTL_SC_TXG:
		retval = ixj_sidac(board, *(int *)arg);
		break;
	case IXJCTL_AEC_START:          /* enable echo cancellation.	*/
		ixj_aec_start(board, *(int *)arg);
		break;
	case IXJCTL_AEC_STOP:		/* disable echo cancellation.	*/
		aec_stop(board);
		break;
	case IXJCTL_AEC_GET_LEVEL:	/* check current setting of aec.*/
		retval = j->aec_level;
		break;
	case PHONE_PLAY_CODEC:		/* set playback channel codec.	*/
		retval = set_play_codec(board, *(int *)arg);
		break;
	case PHONE_PLAY_START:		/* start DSP play mode.		*/
		retval = ixj_play_start(board);
		break;
	case PHONE_PLAY_STOP:		/* stop DSP play mode.		*/
		ixj_play_stop(board);
		break;
	case PHONE_PLAY_DEPTH:		/* set number of frames DSP will*/
					/* queue. This affects audio	*/
					/* latency. Set to min val of 1	*/
					/* to 2 to reduce latency.	*/
		set_play_depth(board, *(int *)arg);
		break;
	case PHONE_PLAY_VOLUME:		/* set audio volume scaling 	*/
					/* factor for DSP.		*/
		if(*(int *)arg == -1) { /* if -1, return current setting*/
			retval = get_play_volume(board);
		}
		else {                  /* otherwise set value.		*/
			set_play_volume(board, *(int *)arg);
			retval = *(int *)arg;
		}
		break;
	case PHONE_PLAY_LEVEL:          /* check current play level.	*/
		retval = get_play_level(board);
		break;
	case IXJCTL_DSP_TYPE:		/* check DSP type.		*/
		retval = (j->dsp.high << 8) + j->dsp.low;
		break;
	case IXJCTL_DSP_VERSION:        /* get DSP firmware version.	*/
		retval = (j->ver.high << 8) + j->ver.low;
		break;
	case IXJCTL_HZ:			/* tell driver thr interrupt 	*/
					/* rate that kernel was compiled*/
					/* for. only needed if not the 	*/
					/* default val of 100 Hz.	*/
		hertz = *(int *)arg;
		break;
	case IXJCTL_RATE:		/* set poll rate of driver.	*/
		if (*(int *)arg > hertz)
			retval = -1;
		else
			samplerate = *(int *)arg;
		break;
	case IXJCTL_DRYBUFFER_READ:	/* returns number of buffer	*/
					/* under-runs that have occured.*/
		*(unsigned long *)arg = j->drybuffer;
		break;
	case IXJCTL_DRYBUFFER_CLEAR:    /* clear the drybuffer counter.	*/
		j->drybuffer = 0;
		break;
	case IXJCTL_FRAMES_READ:        /* returns number of frames	*/
					/* read by DSP since driver was	*/
					/* opened.			*/
		*(unsigned long *)arg = j->framesread;
		break;
	case IXJCTL_FRAMES_WRITTEN:	/* returns number of frames	*/
					/* written by DSP since driver	*/
					/* was opened.			*/
		*(unsigned long *)arg = j->frameswritten;
		break;
	case IXJCTL_READ_WAIT:		/* returns number of times that	*/
					/* read was blocked if opened	*/
					/* in nonblocking mode.		*/
		*(unsigned long *)arg = j->read_wait;
		break;
	case IXJCTL_WRITE_WAIT:         /* returns number of times that	*/
					/* write was blocked if opened	*/
					/* in nonblocking mode.		*/
		*(unsigned long *)arg = j->write_wait;
		break;
	case PHONE_MAXRINGS:		/* set maximum number of rings	*/
		j->maxrings = *(char *)arg;					
		break;
	case PHONE_SET_TONE_ON_TIME:	/* set call progress tone	*/
					/* generation "on" time.	*/
		ixj_set_tone_on(*(int *)arg, board);
		break;
	case PHONE_SET_TONE_OFF_TIME:	/* set call progress tone	*/
					/* generation "off" time.	*/
		ixj_set_tone_off(*(int *)arg, board);
		break;
	case PHONE_GET_TONE_ON_TIME:    /* get tone "on" time setting.	*/
		if (ixj_get_tone_on(board)) {
			retval = -1;
		} else {
			retval = (j->ssr.high << 8) + j->ssr.low;
		}
		break;
	case PHONE_GET_TONE_OFF_TIME:	/* get tone "off" time setting.	*/
		if (ixj_get_tone_off(board)) {
			retval = -1;
		} else {
			retval = (j->ssr.high << 8) + j->ssr.low;
		}
		break;
	case PHONE_PLAY_TONE:		/* play selected tone. tone	*/
					/* index passed as arg.		*/
		if (!j->tone_state)
			retval = ixj_play_tone(board, *(char *)arg);
		else
			retval = -1;
		break;
	case PHONE_GET_TONE_STATE:	/* determine if tone is being	*/
					/* played.			*/
		retval = j->tone_state;
		break;
	case PHONE_DTMF_READY:		/* poll to determine if DTMF dig*/
					/* was pressed.			*/
		retval = j->ex.bits.dtmf_ready;
		break;
	case PHONE_GET_DTMF:            /* get next DTMF digit in buffer*/
					/* returns number representing	*/
					/* the digit.			*/
		if (ixj_hookstate(board)) {
			if (j->dtmf_rp != j->dtmf_wp) {
				retval = j->dtmfbuffer[j->dtmf_rp];
				j->dtmf_rp++;
				if (j->dtmf_rp == 79)
					j->dtmf_rp = 0;
				if (j->dtmf_rp == j->dtmf_wp) {
					j->ex.bits.dtmf_ready = j->dtmf_rp = j->dtmf_wp = 0;
				}
			}
		}
		break;
	case PHONE_GET_DTMF_ASCII:     	/* get next DTMF digit in buffer*/
					/* returns char representing	*/
					/* the digit.			*/
		if (ixj_hookstate(board)) {
			if (j->dtmf_rp != j->dtmf_wp) {
				switch (j->dtmfbuffer[j->dtmf_rp]) {
				case 10:
					retval = 42;	/*'*'; 	*/
					break;
				case 11:
					retval = 48;	/*'0';	*/
					break;
				case 12:
					retval = 35;	/*'#';	*/
					break;
				case 28:
					retval = 65;	/*'A';	*/
					break;
				case 29:
					retval = 66;	/*'B';	*/
					break;
				case 30:
					retval = 67;	/*'C';	*/
					break;
				case 31:
					retval = 68;	/*'D';	*/
					break;
				default:
					retval = 48 + j->dtmfbuffer[j->dtmf_rp];
					break;
				}
				j->dtmf_rp++;
				if (j->dtmf_rp == 79)
					j->dtmf_rp = 0;
				if(j->dtmf_rp == j->dtmf_wp)
				{
					j->ex.bits.dtmf_ready = j->dtmf_rp = j->dtmf_wp = 0;
				}
			}
		}
		break;
	case PHONE_DTMF_OOB:		/* send DTMF only out of band.	*/
		j->flags.dtmf_oob = *(int *)arg;
		break;
	case PHONE_DIALTONE:		/* play standard US dialtone.	*/ 
		ixj_dialtone(board);
		break;
	case PHONE_BUSY:		/* play standard US busytone.	*/
		ixj_busytone(board);
		break;
	case PHONE_RINGBACK:		/* play stand US ringback tone	*/
		ixj_ringback(board);
		break;
	case PHONE_WINK:
		if(j->cardtype == QTI_PHONEJACK) {
			retval = -1;
		}
		else {
			retval = ixj_wink(board);
		}
		break;		
	case PHONE_CPT_STOP:		/* stop playing current tone	*/ 
		ixj_cpt_stop(board);
		break;
        case PHONE_QUERY_CODEC:
        {
                struct phone_codec_data pd;
                int val;
                int proto_size[] = {
                        -1,
                        12, 10, 16, 9, 8, 48, 5,
                        40, 40, 80, 40, 40, 6
                };
		pd = *(struct phone_codec_data *)arg;
                if(pd.type<1 || pd.type>13)
                        return EPROTONOSUPPORT;
                if(pd.type<G729)
                        val=proto_size[pd.type];
                else switch(j->baseframe.low)
                {
                        case 0xA0:val=2*proto_size[pd.type];break;
                        case 0x50:val=proto_size[pd.type];break;
                        default:val=proto_size[pd.type]*3;break;
                }
                pd.buf_min=pd.buf_max=pd.buf_opt=val;
		*(struct phone_codec_data *)arg = pd;
                return 0;
        }
	case IXJCTL_DSP_IDLE:		/* put DSP in idle mode.	*/
		idle(board);
		break;
	case IXJCTL_MIXER:		/* set mixer values.		*/
		if ((*(long *)arg & 0xff) == 0xff)
			retval = ixj_get_mixer(*(long *)arg,board);
		else
			ixj_mixer(*(long *)arg, board);
		break;
	case IXJCTL_DAA_COEFF_SET:	/* select country-of-origin	*/
					/* setting of the PSTN CO.	*/
		switch (*(int *)arg) {
		case DAA_US:
			DAA_Coeff_US(board);
			retval = ixj_daa_write(board);
			break;
		case DAA_UK:
			DAA_Coeff_UK(board);
			retval = ixj_daa_write(board);
			break;
		case DAA_FRANCE:
			DAA_Coeff_France(board);
			retval = ixj_daa_write(board);
			break;
		case DAA_GERMANY:
			DAA_Coeff_Germany(board);
			retval = ixj_daa_write(board);
			break;
		case DAA_AUSTRALIA:
			DAA_Coeff_Australia(board);
			retval = ixj_daa_write(board);
			break;
		case DAA_JAPAN:
			DAA_Coeff_Japan(board);
			retval = ixj_daa_write(board);
			break;
		default:
			retval = 1;
			break;
		}
		break;
	case IXJCTL_DAA_AGAIN:		/* set analog gain of DAA amps	*/
					/* on PSTN port.		*/
		ixj_daa_cr4(board, *(char *)arg | 0x02);
		break;
	case IXJCTL_PSTN_LINETEST:	/* test the PSTN port on	*/
					/* LineJack.			*/
		retval = ixj_linetest(board);
		break;
	case IXJCTL_VMWI:		/* turn on visual message	*/
					/* waiting indicator.		*/
		ixj_write_vmwi(board, *(int *)arg);
		break;
	case IXJCTL_CID:		/* get CID data.		*/
		*(PHONE_CID *)arg = j->cid;
		j->ex.bits.caller_id = 0;
		break;
	case IXJCTL_WINK_DURATION:	/* set min wink duration in ms	*/
		j->winktime = *(int *)arg;
		break;
	case IXJCTL_PORT:		/* set which port to use for 	*/
					/* audio.			*/
		if (*(int *)arg) {
			retval = ixj_set_port(board, *(int *)arg);
		} else {
			retval = j->port;
		}
		break;
	case IXJCTL_POTS_PSTN:		/* connect POTS to PSTN port on */
					/* linejack.			*/
		retval = ixj_set_pots(board, *(int *)arg);
		break;
	case PHONE_CAPABILITIES:	/* returns number of		*/
					/* capabilities on card.	*/
		retval = j->caps;
		break;
	case PHONE_CAPABILITIES_LIST:	/* fill passed struct with this	*/
					/* card's capabilities.		*/
		memcpy((char *)arg, j->caplist, sizeof(struct phone_capability) * j->caps);
		break;
	case PHONE_CAPABILITIES_CHECK:	/* fill passed struct with this	*/
					/* card's capabilities.		*/
		retval = capabilities_check(board, (struct phone_capability *) arg);
		break;
	case PHONE_PSTN_SET_STATE:	/* set mode of DAA...		*/
		daa_set_mode(board, *(int *)arg);
		break;
	case PHONE_PSTN_GET_STATE:	/* get state of DAA...		*/
		retval = j->daa_mode;
		j->ex.bits.pstn_ring = 0;
		break;
	case IXJCTL_SET_FILTER:		/* select and program tone 	*/
					/* detection filter.		*/
		jf = *(IXJ_FILTER *)arg;
		retval = ixj_init_filter(board, &jf);
		break;
	case IXJCTL_GET_FILTER_HIST:	/* get selected filter history.	*/
		retval = j->filter_hist[*(short *)arg];
		break;
	case IXJCTL_INIT_TONE:		/* change tone table element.	*/
		ti = *(IXJ_TONE *)arg;
		retval = ixj_init_tone(board, &ti);
		break;
	case IXJCTL_TONE_CADENCE:	/* play tone cadence.		*/
		retval = ixj_build_cadence(board, (IXJ_CADENCE *) arg);
		break;
	case IXJCTL_FILTER_CADENCE:	/* filter cadence...?		*/
		retval = ixj_build_filter_cadence(board, (IXJ_FILTER_CADENCE *) arg);
		break;
	case IXJCTL_SIGCTL:
		j->sigdef = *(IXJ_SIGDEF *)arg;
		j->ixj_signals[j->sigdef.event] = j->sigdef.signal;
		if(j->sigdef.event < 33) {
			raise = 1;
			for(mant = 0; mant < j->sigdef.event; mant++){
				raise *= 2;
			}
			if(j->sigdef.signal)
				j->ex_sig.bytes |= raise; 
			else
				j->ex_sig.bytes &= (raise^0xffff); 
		}
		break;
	case IXJCTL_INTERCOM_STOP:	/* disable intercom.		*/
		j->intercom = -1;
		ixj[*(int *)arg].intercom = -1;
		ixj_record_stop(board);
		ixj_record_stop(*(int *)arg);
		ixj_play_stop(board);
		ixj_play_stop(*(int *)arg);
		idle(board);
		idle(*(int *)arg);
		break;
	case IXJCTL_INTERCOM_START:	/* enable intercom.		*/
		j->intercom = *(int *)arg;
		ixj[*(int *)arg].intercom = board;
		ixj_play_start(*(int *)arg);
		ixj_record_start(board);
		ixj_play_start(board);
		ixj_record_start(*(int *)arg);
		idle(board);
		idle(*(int *)arg);
		break;
	default:
		return EINVAL;
	}
	clear_bit(board, &j->busyflags);

	/* Do some FreeBSD magic to pass back retval to the userland program
	 *
	 * If retval is zero, return 0 (good). The ioctl will then return 0 to
	 * the userland program.
	 *
	 * If retval is positive, we have a value to pass back. So copy the
	 * value into p_retval[0] and return 0 (good). The ioctl will then
	 * return retval to the userland program.
	 *
	 * If retval is negative, it means we had an error
	 * So, simply call return with a positive error code.
	 * The ioctl will then return -1 (indicating failure) to the userland
	 * program and the error will be in the errno variable.
	 */

	if (retval == 0) return 0;
	if (retval > 0) {
		p->p_retval[0] = retval;
		return 0;
	}
	return EINVAL;
}

static int
ixj_poll(dev_t dev, int events, struct proc *p)
{
	IXJ *j = &ixj[UNIT(dev)];
	int revents = 0;
	int s;

	s = spltty();		
	if (events & (POLLIN | POLLRDNORM)) {
		if (j->read_buffer_ready > 0)
			revents |= events & (POLLIN | POLLRDNORM); /* readable 	*/
		else
			selrecord(p, &j->poll_q);
	}

	if(events & (POLLOUT | POLLWRNORM)) {
		if (j->write_buffers_empty > 0)
			revents |= events & (POLLOUT | POLLWRNORM);/* writable 	*/
		else
			selrecord(p, &j->poll_q);
	}
	
	if(events & POLLRDBAND) {	
		if (j->ex.bytes)
			revents |= events & POLLRDBAND;		   /* exception	*/
		else
			selrecord(p, &j->poll_q);
	}
	splx(s);
	return revents;
}

int ixj_WriteDSPCommand(unsigned short cmd, int board)
{
	BYTES bytes;			
	int s;
	int timer = 0;

	s = splhigh();				/* disable interrupts	*/	
	bytes.high = (cmd & 0xFF00) >> 8;
	bytes.low = cmd & 0x00FF;
	
	while (!IsControlReady(board)) { 	/* wait up to 60ms	*/
		if(timer >= 12000) {
			printf("ixj%d: DSP Control Ready Fail\n", board);
			return (-1);
		}
		DELAY(5);
		timer++;
	}
	
	outb_p(bytes.low, ixj[board].DSPbase + 6);
	outb_p(bytes.high, ixj[board].DSPbase + 7);

	if (ixj_status_wait(board)) {
		ixj[board].ssr.low = 0xFF;
		ixj[board].ssr.high = 0xFF;
		return -1;
	}
					/* Read Software Status Register */
	ixj[board].ssr.low = inb_p(ixj[board].DSPbase + 2);
	ixj[board].ssr.high = inb_p(ixj[board].DSPbase + 3);

	splx(s);			/* return normal intr level	*/
	return 0;
}


int ixj_status_wait(int board)
{
	int timer = 0;

	while (!IsStatusReady(board)) {		/* wait up to 60ms	*/
		if(timer >= 12000)
			return (-1);
		DELAY(5);
		timer++;
	}
	return 0;
}

int ixj_PCcontrol_wait(int board)
{
	int timer = 0;
	
	while (!IsPCControlReady(board)) { 	/* wait up to 60ms	*/
		if(timer >= 12000)
			return (-1);
		DELAY(5);
		timer++;
	}
	return 0;
}

						/* dummy functs for now		*/
void clear_bit(int nr, void *addr)
{
	return;
}

int test_and_set_bit(int nr, void *addr)
{
	return 0;
}
