/******************************************************************************
 * (c) Copyright 2000, Devin Butterfield, Robert A. Buaas
 * 
 * ixjvar_bsd.h	
 *
 * 		This file contains the device softc, global variables, defines, 
 * 		and function prototypes for the FreeBSD driver.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	Devin Butterfield, <db@db.wireless.net>
 *		Robert A. Buaas, <buaas@wireless.net>
 *
 * Notes:
 *
 * $Id: ixjvar_bsd.h,v 1.1 2001/03/03 12:18:34 devinb Exp $
 *
 *****************************************************************************/

#define HZ		100		/* cpu clock frequency		*/

#define MAXRINGS 	5

/************************************************************************
*
* ixjdebug meanings are now bit mapped instead of level based
* Values can be or'ed together to turn on multiple messages
*
* bit  0 (0x0001) = any failure
* bit  1 (0x0002) = general messages
* bit  2 (0x0004) = POTS ringing related
* bit  3 (0x0008) = PSTN events
* bit  4 (0x0010) = PSTN Cadence state details
* bit  5 (0x0020) = Tone detection triggers
* bit  6 (0x0040) = Tone detection cadence details
* bit  7 (0x0080) = ioctl tracking
* bit  8 (0x0100) = signal tracking
* bit  9 (0x0200) = CallerID generation details
*
************************************************************************/

#define ixjdebug	0

#define TYPE(dev) 	(MINOR(dev) >> 4)	
#define NUM(dev) 	(MINOR(dev) & 0xf)

#define in_interrupt()	(cpl != 0)	/* to check for intr context	*/ /* XXX: This is likely wrong!*/

#define outb_p(v, a)    outb(a, v)     	/* io_p machine instructions 	*/
#define outw_p(v, a)    outw(a, v)     	/* redefined for FreeBSD      	*/
#define inb_p(a)        inb(a)
#define inw_p(a)        inw(a)

#define udelay(t) 	DELAY(t)	/* redefine delay for FreeBSD	*/

					/* These two defines are from	*/
					/* <linux/timer.h>		*/
#define time_after(a,b) ((long)(b) - (long)(a) < 0)
#define time_before(a,b) time_after(b,a)

					/* hack to allow the use of	*/
					/* linux's printk()		*/
#define KERN_INFO
#define printk		printf

#define IXJ_STATE_OPEN	0x0001		/* flag indicating dev is opened*/

struct ixj_softc {              	/* param struct per a dev.	*/
	dev_t		dev;            /* basic device structure	*/
	int		dsp_rid;	/* resource id for dsp	 	*/
	int		xilinx_rid;	/* resource id for xilinx	*/
        struct resource	*dsp_ioport;	/* resource for dsp port 	*/
	struct resource *xilinx_ioport;	/* resource for xilinx port	*/
	int		DSPbase;        /* base address of DSP		*/
	int 		XILINXbase;     /* base address of XILINX	*/
	struct proc	*sc_async;	/* process that wants SIGIO	*/
};

					/* function prototypes		*/

int	ixj_attach		__P((int, struct ixj_softc *));
int	ixj_status_wait		__P((int));
int	ixj_WriteDSPCommand	__P((unsigned short, int));
void	ixj_add_timer		__P((void));
void 	ixj_timeout		__P((void *));
int 	ixj_PCcontrol_wait	__P((int));
int 	set_base_frame		__P((int, int));
int 	set_play_codec		__P((int, int));
char 	daa_get_version		__P((int));
int 	ixj_daa_cid_reset	__P((int));
BOOL 	SLIC_SetState		__P((BYTE, int));
int 	ixj_mixer		__P((long, int));
void 	set_rec_depth		__P((int, int));
void 	add_caps		__P((int));
int 	Stub			__P((IXJ *, unsigned long));
int 	capabilities_check	__P((int, struct phone_capability *));
void 	ixj_testram		__P((int));
int 	ixj_get_tone_off	__P((int));
int 	ixj_daa_cr4		__P((int, char));
void 	ixj_aec_start		__P((int, int));
int 	get_rec_level		__P((int));
int 	get_rec_volume		__P((int));
int 	set_rec_codec		__P((int, int));	
void 	ixj_write_vmwi		__P((int, int));
void 	ixj_ring_start		__P((int));
int 	ixj_ring		__P((int));
void 	ixj_write_cidcw		__P((int));	
void	ixj_write_cid		__P((int));
int 	ixj_set_port		__P((int, int));
int 	ixj_set_pots		__P((int, int));
int 	ixj_hookstate		__P((int));
int 	ixj_record_start	__P((int));
void 	ixj_record_stop		__P((int));
void 	set_rec_volume		__P((int, int));
void	ixj_vad			__P((int, int));
int 	ixj_play_start		__P((int));
void	ixj_play_stop		__P((int));
int 	ixj_set_tone_on		__P((unsigned short, int));
int 	ixj_set_tone_off	__P((unsigned short, int));
int 	ixj_get_tone_on		__P((int));
int 	ixj_play_tone		__P((int, char));
int 	idle			__P((int));
void 	ixj_ring_off		__P((int));
void 	aec_stop		__P((int));
void 	ixj_ringback		__P((int));
void 	ixj_busytone		__P((int));
void 	ixj_dialtone		__P((int));
void 	ixj_cpt_stop		__P((int));
int 	daa_set_mode		__P((int, int));
int 	ixj_linetest		__P((int));
int 	ixj_daa_write		__P((int));
void 	DAA_Coeff_US		__P((int));
void 	DAA_Coeff_UK		__P((int));
void 	DAA_Coeff_France	__P((int));
void 	DAA_Coeff_Germany	__P((int));
void 	DAA_Coeff_Australia	__P((int));
void	DAA_Coeff_Japan		__P((int));
int	ixj_wink		__P((int));
void	clear_bit		__P((int,void *)); /* XXX NOP FUNCTION 	*/
int	test_and_set_bit	__P((int,void *)); /* XXX NOP FUNCTION	*/
int 	ixj_init_filter		__P((int, IXJ_FILTER *));
int 	ixj_init_tone		__P((int, IXJ_TONE *));
int 	ixj_build_cadence	__P((int, IXJ_CADENCE *));
void 	ulaw2alaw		__P((unsigned char *, unsigned long));
void 	alaw2ulaw		__P((unsigned char *, unsigned long));
void 	set_dtmf_prescale	__P((int, int));
int 	get_dtmf_prescale	__P((int));
int 	ixj_get_mixer		__P((long, int));
int 	ixj_siadc		__P((int, int));
int 	ixj_sidac		__P((int, int));
int 	ixj_build_filter_cadence		__P((int, IXJ_FILTER_CADENCE *));
extern __inline__ void 	LED_SetState		__P((int, int));
extern __inline__ void 	ixj_read_HSR		__P((int));
extern __inline__ int 	IsControlReady		__P((int));
extern __inline__ int 	IsPCControlReady	__P((int));
extern __inline__ int 	IsStatusReady		__P((int));
extern __inline__ int 	IsRxReady		__P((int));
extern __inline__ int 	IsTxReady		__P((int));
extern __inline__ void 	set_play_volume		__P((int, int));
extern __inline__ int 	get_play_volume		__P((int));
extern __inline__ BYTE 	SLIC_GetState		__P((int));
extern __inline__ int 	ixj_gpio_read		__P((int));
extern __inline__ int 	get_play_level		__P((int));
extern __inline__ void 	set_play_depth		__P((int, int));
extern __inline__ void 	ixj_kill_fasync		__P((int, IXJ_SIGEVENT));


					/* global variables/functs	*/
#ifdef __COMMON__ 

/************************************************************************
*
* These are function definitions to allow external modules to register
* enhanced functionality call backs.
*
************************************************************************/

IXJ_REGFUNC ixj_DownloadG729 = &Stub;
IXJ_REGFUNC ixj_DownloadTS85 = &Stub;
IXJ_REGFUNC ixj_PreRead = &Stub;
IXJ_REGFUNC ixj_PostRead = &Stub;
IXJ_REGFUNC ixj_PreWrite = &Stub;
IXJ_REGFUNC ixj_PostWrite = &Stub;
IXJ_REGFUNC ixj_PreIoctl = &Stub;
IXJ_REGFUNC ixj_PostIoctl = &Stub;
int samplerate = 100;
int hertz = HZ;
IXJ ixj[IXJMAX];
struct callout_handle ch;
long jifvar;				/* var for jiffies hack		*/
#else
extern IXJ_REGFUNC ixj_DownloadG729;
extern IXJ_REGFUNC ixj_DownloadTS85;
extern IXJ_REGFUNC ixj_PreRead;
extern IXJ_REGFUNC ixj_PostRead;
extern IXJ_REGFUNC ixj_PreWrite;
extern IXJ_REGFUNC ixj_PostWrite;
extern IXJ_REGFUNC ixj_PreIoctl;
extern IXJ_REGFUNC ixj_PostIoctl;
extern int samplerate;			/* driver samplerate		*/
extern int hertz;			/* system cpu clock rate	*/
extern IXJ ixj[IXJMAX];
extern struct callout_handle ch;	/* handle for timeout()		*/
extern long jifvar;			/* var for jiffnes hack		*/
#endif

					/* hack to make jiffies		*/
					/* long int format...		*/
#define	jiffies		(jifvar = ticks)

