/******************************************************************************
 * 
 * ixjvar_linux.h	
 *
 * 		This file contains the global variables, defines, 
 * 		and function prototypes for the linux driver.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	Devin Butterfield, <dbutter@wireless.net>
 *
 * Notes:
 *
 * $Id: ixjvar_linux.h,v 1.1 2001/03/03 12:18:34 devinb Exp $
 *
 *****************************************************************************/

#define PERFMON_STATS
#define IXJDEBUG 0
#define MAXRINGS 5

/************************************************************************
*
* ixjdebug meanings are now bit mapped instead of level based
* Values can be or'ed together to turn on multiple messages
*
* bit  0 (0x0001) = any failure
* bit  1 (0x0002) = general messages
* bit  2 (0x0004) = POTS ringing related
* bit  3 (0x0008) = PSTN events
* bit  4 (0x0010) = PSTN Cadence state details
* bit  5 (0x0020) = Tone detection triggers
* bit  6 (0x0040) = Tone detection cadence details
* bit  7 (0x0080) = ioctl tracking
* bit  8 (0x0100) = signal tracking
* bit  9 (0x0200) = CallerID generation details
*
************************************************************************/

void ixj_add_timer(void);
void ixj_timeout(unsigned long ptr);
int ixj_fasync(int fd, struct file *, int mode);
int ixj_set_port(int board, int arg);
int ixj_set_pots(int board, int arg);
int ixj_hookstate(int board);
int ixj_record_start(int board);
void ixj_record_stop(int board);
void set_rec_volume(int board, int volume);
int get_rec_volume(int board);
int set_rec_codec(int board, int rate);
void ixj_vad(int board, int arg);
int ixj_play_start(int board);
void ixj_play_stop(int board);
int ixj_set_tone_on(unsigned short arg, int board);
int ixj_set_tone_off(unsigned short, int board);
int ixj_play_tone(int board, char tone);
void ixj_aec_start(int board, int level);
int idle(int board);
void ixj_ring_off(int board);
void aec_stop(int board);
void ixj_ringback(int board);
void ixj_busytone(int board);
void ixj_dialtone(int board);
void ixj_cpt_stop(int board);
int daa_set_mode(int board, int mode);
int ixj_daa_write(int board);
void DAA_Coeff_US(int board);
void DAA_Coeff_UK(int board);
void DAA_Coeff_France(int board);
void DAA_Coeff_Germany(int board);
void DAA_Coeff_Australia(int board);
void DAA_Coeff_Japan(int board);
int ixj_init_filter(int board, IXJ_FILTER * jf);
int ixj_init_tone(int board, IXJ_TONE * ti);
int ixj_build_cadence(int board, IXJ_CADENCE * cp);
int ixj_build_filter_cadence(int board, IXJ_FILTER_CADENCE * cp);
BOOL SLIC_SetState(BYTE byState, int board);
void ulaw2alaw(unsigned char *, unsigned long);
void alaw2ulaw(unsigned char *, unsigned long);
int ixj_PCcontrol_wait(int board);
void ixj_timeout(unsigned long ptr);
void set_dtmf_prescale(int, int);
int get_dtmf_prescale(int);
int ixj_siadc(int, int);
int ixj_sidac(int, int);
void add_caps(int board);
int get_rec_level(int board);
void ixj_testram(int board);
void ixj_write_cidcw(int board);
void ixj_ring_start(int board);
int ixj_ring(int board);
int ixj_get_tone_on(int board);
int ixj_get_tone_off(int board);
int ixj_wink(int board);
int ixj_get_mixer(long val, int board);
int ixj_mixer(long val, int board);
int ixj_daa_cr4(int board, char reg);
int ixj_linetest(int board);
void ixj_write_vmwi(int board, int msg);
char daa_get_version(int board);
int ixj_daa_cid_reset(int board);
void ixj_write_cid(int board);
int set_base_frame(int board, int size);
int set_play_codec(int board, int rate);
void set_rec_depth(int board, int depth);
int Stub(IXJ *, unsigned long);

#ifdef __COMMON__

/************************************************************************
*
* These are function definitions to allow external modules to register
* enhanced functionality call backs.
*
************************************************************************/

IXJ_REGFUNC ixj_DownloadG729 = &Stub;
IXJ_REGFUNC ixj_DownloadTS85 = &Stub;
IXJ_REGFUNC ixj_PreRead = &Stub;
IXJ_REGFUNC ixj_PostRead = &Stub;
IXJ_REGFUNC ixj_PreWrite = &Stub;
IXJ_REGFUNC ixj_PostWrite = &Stub;
IXJ_REGFUNC ixj_PreIoctl = &Stub;
IXJ_REGFUNC ixj_PostIoctl = &Stub;
IXJ ixj[IXJMAX];
int ixjdebug = 0;
int hertz = HZ;
int samplerate = 100;
struct timer_list ixj_timer;
#else
extern IXJ_REGFUNC ixj_DownloadG729;
extern IXJ_REGFUNC ixj_DownloadTS85;
extern IXJ_REGFUNC ixj_PreRead;
extern IXJ_REGFUNC ixj_PostRead;
extern IXJ_REGFUNC ixj_PreWrite;
extern IXJ_REGFUNC ixj_PostWrite;
extern IXJ_REGFUNC ixj_PreIoctl;
extern IXJ_REGFUNC ixj_PostIoctl;
extern IXJ ixj[IXJMAX];
extern int ixjdebug;
extern int hertz;
extern int samplerate;
extern struct timer_list ixj_timer;
#endif

